#include frex:shaders/api/accessibility.glsl
#include frex:shaders/api/fog.glsl
#include frex:shaders/api/light.glsl
#include frex:shaders/api/material.glsl
#include frex:shaders/api/player.glsl
#include frex:shaders/api/view.glsl
#include frex:shaders/api/world.glsl
#include frex:shaders/lib/math.glsl
#include ecos:shaders/lib/tonemap.glsl
#include ecos:shaders/lib/shadow.glsl
#include ecos:stylize_options

#define GODRAYS

uniform sampler2D glint_item;
uniform sampler2D glint_entity;
uniform sampler2D colored_lights;

in float v_diffuse;
in float v_heldCosIn;
in float v_heldCosOut;

#ifdef SHADOW_MAP_PRESENT
in vec4 v_shadowViewPos;
#endif

out vec4 fragColor;
out vec4 fragBloom;
out vec4 fragNormal;

#ifdef SHADOW_MAP_PRESENT
// Shadow functions
vec3 shadowDist(int cascade)
{
  vec4 c = frx_shadowCenter(cascade);
  return abs((c.xyz - v_shadowViewPos.xyz) / c.w);
}

int selectShadowCascade()
{
  vec3 d3 = shadowDist(3);
  vec3 d2 = shadowDist(2);
  vec3 d1 = shadowDist(1);

  int cascade = 0;

  if (all(lessThan(d3, vec3(1.0))))
  {
    cascade = 3;
  }
  else if (all(lessThan(d2, vec3(1.0))))
  {
    cascade = 2;
  } 
  else if (all(lessThan(d1, vec3(1.0))))
  {
    cascade = 1;
  }

  return cascade;
}
// End shadow functions
#endif

vec3 getGlint()
{
  // vanilla scale factor for entity, works in most scenario
  const float scale = 0.16;

  // vanilla translation factor
  float time = frx_renderSeconds * 8. * frx_glintSpeed;
  vec2 translation = vec2(-time / 110., time / 30.);
  vec2 uv = frx_normalizeMappedUV(frx_texcoord) * scale + translation;
  vec3 glint = mix(texture(glint_item, uv).rgb, texture(glint_entity, uv).rgb, float(frx_matGlintEntity));

  // emulate GL_SRC_COLOR sfactor
  return glint * glint * frx_glintStrength * float(frx_matGlint);
}

float shiny_glint_noise(vec2 uv, float zoom)
{
    const float skew = 0.4;
    const float detail = 8.0;
    {
      uv += uv.y * skew;
      uv *= zoom;
    }
    
    float t = mod(uv.x, 1.0);
    {
      t *= detail;
    }

    float f = fract(t);
    {
      t -= f;
      t /= detail;
      f -= 0.5;
      f = abs(f);
      f = 1.0 - 2.0 * f;
    }
    
    float n = frx_noise2d(vec2(t)) * f;
    return n * n;
}

vec3 getLight() {
  if (frx_isGui && !frx_isHand)
  {
    return vec3(frx_fragEnableDiffuse ? v_diffuse : 1.0);
  }

  float sqrtTransition = sqrt(frx_skyLightTransitionFactor);
  float sqt = frx_skyLightTransitionFactor * frx_skyLightTransitionFactor;
  float skyStrength = mix(2.0, 0.4, sqt * float(frx_worldIsMoonlit)) * (1.0 - 0.7 * frx_rainGradient - 0.15 * frx_thunderGradient);
  float directSkyStrength = skyStrength * sqrtTransition;
  vec3 night_ambient = mix(vec3(1.0), vec3(0.3, 0.3, 1.0), sqrtTransition * float(frx_worldIsMoonlit));

  float directSkyLight;
  {
    #ifdef SHADOW_MAP_PRESENT
    {
      int cascade = selectShadowCascade();
      vec4 shadowPos = frx_shadowProjectionMatrix(cascade) * v_shadowViewPos;
      vec3 shadowTexCoord = shadowPos.xyz * 0.5 + 0.5;

      // Sample the shadow map
      #if SHADOW_MAP_SIZE < 1024
      {
        directSkyLight = sampleShadowPCF(frxs_shadowMap, shadowTexCoord, float(cascade));
      }
      #else
      {
        directSkyLight = texture(frxs_shadowMap, vec4(shadowTexCoord.xy, float(cascade), shadowTexCoord.z));
      }
      #endif
    }
    #else
    {
      directSkyLight = 1.0;
    }
    #endif

    // Multiply the color once more with the diffuse shading
    if (frx_fragEnableDiffuse)
    {
      directSkyLight = min(directSkyLight, v_diffuse);
    }

    directSkyLight *= directSkyStrength;
  }

  vec3 block_light;
  {
    block_light = vec3(0.0);

    #ifdef VANILLA_LIGHTING
    {
      block_light = vec3(frx_fragLight.x);
    }
    #endif
    
    #ifdef COLORED_LIGHTS_ENABLED
    {
      vec4 colored_light = frx_getLightFiltered(colored_lights, (frx_vertex.xyz + frx_vertexNormal * 0.1) * (1.0 - float(frx_isHand)) + frx_cameraPos);
      block_light = mix(block_light, colored_light.rgb, colored_light.a);
    }
    #endif

    block_light = block_light * block_light  * 0.4 + pow(block_light, vec3(5.0)) * 0.6;
    block_light *= (3.0 - frx_eyeBrightness.y * directSkyStrength);
  }

  vec3 sun_light;
  {
    sun_light = vec3(1.0, sqt, sqt * sqt) * directSkyLight; 
    sun_light *= night_ambient;
    
    #ifdef VANILLA_LIGHTING
    {
      sun_light *= frx_fragLight.y;
    }
    #endif
  }

  vec3 ambient_sky_light;
  {
    ambient_sky_light = vec3(0.4 * skyStrength);
    ambient_sky_light *= night_ambient;

     #ifdef VANILLA_LIGHTING
    {
      ambient_sky_light *= frx_fragLight.y;
    }
    #endif
  }

  vec3 heldLight;
  {
    if (frx_heldLight.w > 0.0)
    {
      float dist = 1.0 - min(frx_distance / (frx_heldLight.w * 14.0), 1.0);
      float angle = dot(frx_cameraView, frx_vertex.xyz / frx_distance);
      float fDir = smoothstep(v_heldCosOut, v_heldCosIn, angle);

      heldLight = frx_heldLight.rgb * fDir * dist;
      heldLight = heldLight * heldLight;
    }
    else
    {
      heldLight = vec3(0.0);
    }
  }

  vec3 ambient_light = vec3(0.04 + 0.3 - float(frx_worldHasSkylight) * 0.3);

  vec3 light_color;
  {
    light_color = block_light + sun_light + ambient_sky_light + heldLight + ambient_light;
    light_color *= frx_fragEnableAo ? frx_fragLight.z * frx_fragLight.z : 1.0;

    // Apply emissive (yes we apply it at bloom as well)
    light_color += vec3(frx_fragEmissive);

    // Soften strong lights
    light_color = sqrt(light_color * (0.25 + 0.25 * frx_viewBrightness));
  }

  return light_color;
}

void frx_pipelineFragment()
{
  vec4 color = frx_fragColor;

  if (color.a <= 0.0) {
    discard;
  }

  vec3 light = getLight();
  float bloom = 0.0;

  {
    color.rgb *= light;
    color.rgb += getGlint();
  }
  

  if (!frx_isGui || frx_isHand) {
    {
      color.rgb += vec3(float(frx_matFlash));
      color.rgb = mix(color.rgb, vec3(1.0, 0.0, 0.0), float(frx_matHurt) * 0.25);
    }

    #ifdef PBR_ENABLED
    {
      mat3 TBN = mat3(frx_vertexTangent.xyz, cross(frx_vertexTangent.xyz, frx_vertexNormal) * frx_vertexTangent.w, frx_vertexNormal);
      frx_fragNormal = TBN * frx_fragNormal;

      vec3 fragView = -frx_vertex.xyz / frx_distance;
      vec3 reflection = reflect(fragView, frx_fragNormal);
      reflection = TBN * reflection;
      reflection *= sign(reflection.z);

      vec2 shiny_glint_uv = reflection.xy / (reflection.z + 0.000001);
      float alpha = 1.0 - frx_fragRoughness;
      float alpha2 = alpha * alpha;
      float shiny_glint = shiny_glint_noise(shiny_glint_uv, 0.4 * alpha2) * alpha2 * reflection.z;

      color.rgb += mix(0.5 * light * shiny_glint, (shiny_glint - 0.5 * alpha) * light * frx_fragColor.rgb, alpha * frx_fragReflectance);
      color.a += (1.0 - color.a) * pow(max(0.0, 1.0 - dot(fragView, frx_vertexNormal)), 10.0);
    }
    #endif
    
    // Tonemap for funsies (see footnote 1)
    color.rgb = tonemap(color.rgb);

    bloom = frx_fragEmissive;
  
    // Fog
    {
      float fogDensity = pow(smoothstep(0.0, 512.0, frx_distance), 0.4) * pow(smoothstep(512.0, -32.0, frx_vertex.y), 4.0);
      float fogFactor = max(fogDensity, smoothstep(frx_fogStart, frx_fogEnd, frx_distance));
      color.rgb = mix(color.rgb, frx_fogColor.rgb, fogFactor);
      bloom = mix(bloom, 0, fogFactor);
    }
  }

  // Output
  {
    fragColor = color;
    fragBloom = vec4(bloom * color.a, 0.0, 0.0, color.a);
    gl_FragDepth = gl_FragCoord.z;
  }

  #ifdef OUTLINE
  {
    if (frx_renderTargetSolid) {
      vec4 pos = frx_viewMatrix * vec4(frx_vertex.xyz, 1.0);

      fragNormal.rgb = frx_normalModelMatrix * frx_vertexNormal;
      fragNormal = vec4(fragNormal.xy * 0.5 + 0.5, sqrt(1.0 / -pos.z), color.a);

      #ifdef CORNER_HIGHLIGHTS
      {
        fragBloom.g = dot(light, vec3(0.3333));
      }
      #endif
    }
  }
  #endif
}

// footnote 1: based on this calculation https://www.desmos.com/calculator/kmolpqoaxl
