#include ecos:shaders/pass/header.glsl
#include ecos:options
#include ecos:shaders/lib/taa.glsl
#include frex:shaders/api/view.glsl
#include frex:shaders/api/world.glsl

uniform sampler2D img_current;
uniform sampler2D img_history;
uniform sampler2D img_depth;

in vec2 texcoord;

out vec4 fragColor;


#define FEEDBACK_MAX 0.9
#define FEEDBACK_MIN 0.7

vec2 inv_size = vec2(1.0) / vec2(frxu_size);

vec4 clipAABB(vec3 colorMin, vec3 colorMax, vec4 currentColor, vec4 previousColor)
{
  vec3 p_clip = 0.5 * (colorMax + colorMin);
  vec3 e_clip = 0.5 * (colorMax - colorMin);
  vec4 v_clip = previousColor - vec4(p_clip, currentColor.a);
  vec3 v_unit = v_clip.rgb / e_clip;
  vec3 a_unit = abs(v_unit);
  float ma_unit = max(a_unit.x, max(a_unit.y, a_unit.z));

  if (ma_unit > 1.0) {
    return vec4(p_clip, currentColor.a) + v_clip / ma_unit;
  } else {
    return previousColor;// point inside aabb
  }
}

vec3 calcPrevNdc(vec3 currentNdc)
{
  vec4 temp = frx_inverseViewProjectionMatrix * vec4(currentNdc, 1.0);
  vec3 currentPos = temp.xyz / temp.w;

  // transform current camera-space pos -> world-space pos (assumed static) -> previous camera-space pos
  vec3 prevPos = (currentPos + frx_cameraPos) - frx_lastCameraPos;

  temp = frx_lastViewProjectionMatrix * vec4(prevPos, 1.0);
  return temp.xyz / temp.w;
}

vec2 calcPrevUv(sampler2D depthBuffer, vec2 currentUv)
{
  vec2 ndcJitter	   = taaJitter(inv_size, frx_renderFrames);
  vec2 prevNdcJitter = taaJitter(inv_size, frx_renderFrames - 1u);

  float depth = texture(depthBuffer, currentUv).r;
  vec2 prevUv = (calcPrevNdc(vec3(currentUv * 2.0 - 1.0 - ndcJitter, depth * 2.0 - 1.0)).xy + prevNdcJitter) * 0.5 + 0.5;

  if (prevUv != clamp(prevUv, 0.0, 1.0)) return currentUv; // out of bounds
  return prevUv;
}

float velocityWeight(sampler2D depthBuffer, vec2 currentUv)
{
  float depth = texture(depthBuffer, currentUv).r;
  vec2 prevUv = calcPrevNdc(vec3(currentUv * 2.0 - 1.0, depth * 2.0 - 1.0)).xy * 0.5 + 0.5;
  return min(1.0, length((currentUv - prevUv) * frxu_size));
}

// based on INSIDE's TAA and https://github.com/ziacko/Temporal-AA
vec4 taa()
{
  const vec2 box[] = vec2[](
    vec2(-1, -1),
    vec2( 0, -1),
    vec2( 1, -1),
    vec2(-1,  0),
    vec2( 1,  0),
    vec2(-1,  1),
    vec2( 0,  1),
    vec2( 1,  1)
  );

  const vec2 plus[] = vec2[](
    vec2(-1, 0),
    vec2(0, -1),
    vec2(1, 0),
    vec2(0, 1)
  );

  vec2 prevCoord	  = calcPrevUv(img_depth, texcoord);
  vec4 currentColor = texture(img_current, texcoord);
  vec4 historyColor = texture(img_history, prevCoord);

  vec3 minColor0 = currentColor.rgb;
  vec3 maxColor0 = currentColor.rgb;
  for(int i = 0; i < 8; i++) {
    vec3 sampled = texture(img_current, texcoord + box[i] * inv_size).rgb;
    minColor0 = min(minColor0, sampled);
    maxColor0 = max(maxColor0, sampled);
  }

  vec3 minColor1 = currentColor.rgb;
  vec3 maxColor1 = currentColor.rgb;
  for(int i = 0; i < 4; i++) {
    vec3 sampled = texture(img_current, texcoord + plus[i] * inv_size).rgb;
    minColor1 = min(minColor1, sampled);
    maxColor1 = max(maxColor1, sampled);
  }

  vec3 mixedMin = mix(minColor0, minColor1, 0.5);
  vec3 mixedMax = mix(maxColor0, maxColor1, 0.5);

  float feedback = mix(FEEDBACK_MAX, FEEDBACK_MIN, velocityWeight(img_depth, texcoord));
  vec4 clippedHistoryColor = clipAABB(mixedMin, mixedMax, currentColor, historyColor);

  return mix(currentColor, clippedHistoryColor, feedback);
}

void main()
{
  #ifdef ANTIALIASING
  {
    if (frx_renderFrames == 0u) {
      fragColor = texture(img_current, texcoord);
    } else {
      fragColor = taa();
    }
  }
  #else
  {
    fragColor = texture(img_current, texcoord);
  }
  #endif
}
